(****************************************************************)
(*                                                              *)
(*    Fichier logo_principal.ml                                 *)
(*    ------------------------                                  *)
(*    dependance : necessite le module logo_base                *)
(*                                                              *)
(*    date : Janvier 2007 (projet CaML)                         *)
(*                                                              *)
(*    auteur : Courreges Adrian                                 *)
(*             1A IN, groupe D                                  *)
(*                                                              *)
(****************************************************************)

(*--------------------------------------------------------------*)
(*    note importante : dans ce fichier sont renseignees les    *)
(*    sources des fonctions, avec un bref commentaire sur leur  *)
(*    role et leur utilisation. De plus amples dtails sur      *)
(*    leur mode de fonctionnement figurent dans le rapport.     *)
(*--------------------------------------------------------------*)



#open "logo_base";;


(*****************************************************************)
(*                                                               *)
(*                       TYPES DE DONNEES                        *)
(*                                                               *)
(*****************************************************************)

(*---------------------------------------------------------------*)
(*                     type des instructions                     *)
(*---------------------------------------------------------------*)

type instruction = 
     ROTATION of expr
    |DEPLACE of expr
    |SAUT of expr
    |COULEUR of expr * expr * expr
    |REMPLIR
    |ARRETREMPLIR
    |APPEL of string * (expr list)
    |SI of test * (instruction list) * (instruction list)  
    |REPETE of expr * (instruction list)
    |LISTE of (instruction list) ;;


(*---------------------------------------------------------------*)
(*                  type des definitions de procedures           *)
(*     contient le nom de la procedure, ses variables, et son    *)
(*     bloc d'instruction.                                       *)
(*---------------------------------------------------------------*)

type definition = PROC of (string * (string list) * (instruction list)) ;;


(*---------------------------------------------------------------*)
(*                       type des programmes                     *)
(*     contient des listes de definition de procedures et le     *)
(*     bloc principal d'instruction                              *)
(*---------------------------------------------------------------*)

type programme = PROG of (definition list) * (instruction list) ;;


(*---------------------------------------------------------------*)
(*                        type environnement                     *)
(*     contient les def de procedures et la valeur des variables *)
(*     /!\Un environnement est propre a une procedure donnee/!\  *)
(*---------------------------------------------------------------*)

type environnement = ENV of ((definition list) * ((string * float) list));;


(*---------------------------------------------------------------*)
(*                     type de l'etat du systeme                 *)
(*---------------------------------------------------------------*)

type etat = {
             abs : float ;  (* abscisse du curseur               *)
             ord : float ;  (* ordonne du curseur               *)
             ang : float ;  (* orientation du curseur en radians *)
             remp : bool ;  (* mode remplissage                  *)
             ox : float ;   (* abscisse d'origine du remplissage *)
             oy : float     (* ordonne d'origine du remplissage *)
             } ;;



(*****************************************************************)
(*                                                               *)
(*                     ANALYSE SYNTAXIQUE                        *)
(*                                                               *)
(*****************************************************************)


(*---------------------------------------------------------------- 
|  fonction lire_unique_arguments : flux_lecture -> expr          |
|                                                                 |
|  semantique : retourne l'expression d'un seul parametre entre   |
|               parentheses (pour ROTATE, MOVE, JUMP, REPEAT).    |
|               declenche une exception si erreur de syntaxe logo.|
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : expr du parametre                                   |
-----------------------------------------------------------------*)

let lire_unique_argument f =
  match (lit_arguments f) with
    
    |t::[] -> t
    
    |_ -> failwith"mauvais nombre d'arguments !" ;;

(*.................................................................
.  Test : pour un flux commenant par (3)                         .
.         #lire_unique_argument f;;                               .
.         - : expr = Const 3.0                                    .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction lire_rgb : flux_lecture -> expr * expr * expr         |
|                                                                 |
|  semantique : retourne le triplet de 3 parametres entre         |
|               parentheses (pour COLOR).                         |
|               declenche une exception si erreur de syntaxe logo.|
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : triplet d'expr des parametres                       |
------------------------------------------------------------------*)

let lire_rgb f =
  match (lit_arguments f) with
    
    |r::g::b::[] -> (r,g,b)
    
    |_ -> failwith"COLOR doit avoir 3 arguments !" ;;

(*.................................................................
. Test : pour un flux commenant par (3,n,21)                     .
.        #lire_rgb f;;                                            .
.        - : expr * expr * expr = Const 3.0, Var "n", Const 21.0  .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction lire_chaine : flux_lecture -> string                  |
|                                                                 |
|  semantique : retourne la chaine correspondant au nom de la     |
|               procedure.                                        |
|               declenche une exception si erreur de syntaxe logo.| 
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : chaine du nom                                       |
------------------------------------------------------------------*)

let lire_chaine f =
  match (lit_mot_cle f) with
      
    |IDENT(chaine) -> chaine
    
    |_ -> failwith"nom de procedure manquant" ;;

(*.................................................................
.  Test : pour un flux commenant par carre                       .
.         #lire_chaine f;;                                        .
.        - : string = "carre"                                     .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction lit_bloc : flux_lecture -> instruction list           |
|                                                                 |
|  semantique : s'utilise apres lecture d'un BEGIN                |
|               retourne la liste d'instruction d'un bloc et lit  |
|               le END final.                                     |
|               declenche une exception si erreur de syntaxe logo.| 
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : liste des instructions                              |
------------------------------------------------------------------*)

let rec lit_bloc f =
  match (lit_mot_cle f) with
    
    |END -> []
    
    |IF -> 
       let t = (lit_condition f) in
       let sivrai = (lit_bloc f) in
       let sifaux = (lit_bloc f) in
       SI(t,sivrai,sifaux)::(lit_bloc f);
    
    |THEN|ELSE|BEGIN -> (lit_bloc f)
    
    |REPEAT -> 
       let n = (lire_unique_argument f)in
       let tete=(REPETE(n,lit_bloc f)) in
       tete::(lit_bloc f);
    
    |CALL -> 
       let nom=(lire_chaine f) 
       and listargmts=(lit_arguments f) in
       APPEL(nom,listargmts)::(lit_bloc f)
    
    |ROTATE -> 
       let e=(lire_unique_argument f) in
       ROTATION(e)::(lit_bloc f)
    
    |MOVE -> 
       let e=(lire_unique_argument f) in
       DEPLACE(e)::(lit_bloc f)
    
    |JUMP -> 
       let e=(lire_unique_argument f) in
       SAUT(e)::(lit_bloc f)
    
    |COLOR ->
       let (r,g,b)=(lire_rgb f) in
       COULEUR(r,g,b)::(lit_bloc f)
    
    |FILL -> REMPLIR::(lit_bloc f)
    
    |NOFILL -> ARRETREMPLIR::(lit_bloc f)

    |_ -> failwith"Une chaine doit etre precedee de CALL" ;;
 

(*.................................................................
.  Test : pour un flux d'un fichier contenant                     .
.           REPEAT (4)                                            .
.           BEGIN                                                 .
.            MOVE   (n)                                           .
.            ROTATE (90)                                          .
.           END                                                   .
.         END                                                     .
.         #lit_bloc f;;                                           .
.         - : instruction list = [REPETE (Const 4.0,              .
.         [DEPLACE (Var "n"); ROTATION (Const 90.0)])]            .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction lit_definition : flux_lecture -> definition           |
|                                                                 |
|  semantique : s'utilise apres lecture d'un DEF                  |
|               retourne la definition de la procedure avec son   |
|               nom, sa liste d'arguments, et son bloc.           |
|               declenche une exception si erreur de syntaxe logo.| 
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : definition                                          |
------------------------------------------------------------------*)

let lit_definition f =
  let nom = (lire_chaine f) in
  let l = (lit_parametres f) in
  if (lit_mot_cle f)<>BEGIN then failwith "oubli de begin en debut de bloc";
  let bloc=(lit_bloc f)in
  PROC(nom,l,bloc);;

(*.................................................................
. Test : pour un flux d'un fichier contenant                      .
.        carre (n)                                                .
.        BEGIN                                                    .
.          REPEAT (4)                                             .
.          BEGIN                                                  .
.           MOVE   (n)                                            .
.           ROTATE (90)                                           .
.          END                                                    .
.        END                                                      .
.                                                                 .
.        #lit_definition f;;                                      .
.        - : definition =                                         .
.        PROC                                                     .
.        ("carre", ["n"],                                         .
.          [REPETE (Const 4.0,                                    .
.                   [DEPLACE (Var "n"); ROTATION (Const 90.0)])]) .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction lit_programme : flux_lecture -> programme             |
|                                                                 |
|  semantique : lit un programme LOGO et retourne le couple forme |
|               de la liste de definitions de procedures et du    |
|               bloc principal du programme.                      |
|               declenche une exception si erreur de syntaxe logo.|
|                                                                 |
|  arguments : f flux de lecture                                  |
|                                                                 |
|  resultat : definition                                          |
------------------------------------------------------------------*)

let rec lit_programme f =
  match (lit_mot_cle f) with
    
    |DEF -> 
       let procedure=(lit_definition f) in
       let (PROG(procs,main))=(lit_programme f)in
       PROG(procedure::procs,main)
    
    |BEGIN -> 
       let main=(lit_bloc f) in
       PROG([],main)

    |_ -> failwith "structure : def de procedures + bloc";;

(*.................................................................
. Test :#lit_programme (ouvre_fichier "carre.logo");;             .
.- : programme =                                                  .
. PROG                                                            .
.  ([PROC                                                         .
.     ("carre", ["n"],                                            .
.      [REPETE (Const 4.0, [DEPLACE (Var "n");                    .
.       ROTATION (Const 90.0)])])],                               .
.   [SAUT (Const 100.0); ROTATION (Const 90.0); SAUT (Const 100.0);
.    ROTATION (Moins (Const 0.0, Const 90.0)); APPEL ("carre",    .
.    [Const 50.0]); ROTATION (Const 180.0);                       .
.    APPEL ("carre", [Const 50.0])])                              .
..................................................................*)




(*****************************************************************)
(*                                                               *)
(*                         EVALUATION                            *)
(*                                                               *)
(*****************************************************************)


(*---------------------------------------------------------------- 
|   fonction rad_of_deg : float -> float                          |
|                                                                 |
|   semantique : convertit les degres en radians                  |
|                                                                 |
|   arguments : float en degres                                   |
|                                                                 |
|   resultat : float en radians                                   |
------------------------------------------------------------------*)

let rad_of_deg a =
  (a /. 180.) *. (acos(-.1.)) ;;

(*.................................................................
. Test : #rad_of_deg 180.;;                                       .
.        - : float = 3.14159265359                                .
..................................................................*)



(*---------------------------------------------------------------- 
|   fonction valeur_de_var : ('a * 'b) list -> 'a -> 'b           |
|                                                                 |
|   semantique : donne la valeur float associee a une chaine      |
|                dans une liste de couple (chaine,valeur)         |
|                                                                 |
|   arguments : (float * string) list                             |
|               un nom de variable de type string                 |
|                                                                 |
|   resultat : float correspondant a la variable                  |
------------------------------------------------------------------*)

let rec valeur_de_var l nom =
  match l with
      
    |[] -> failwith "evaluation d'une variable indefinie"
    
    |(nomvar,valeur)::q when (nomvar=nom) -> valeur
    
    |t::q -> valeur_de_var q nom ;;

(*.................................................................
. Test : #valeur_de_var [("n",3.);("m",7.)] "m";;                 .
.        - : float = 7.0                                          .
..................................................................*)



(*---------------------------------------------------------------- 
|   fonction evalue_expression : environnement -> expr -> float   | 
|                                                                 |
|   semantique : donne la valeur float d'une expression pour      |
|                un environnement donne.                          |
|                                                                 |
|   arguments : env environnement                                 |
|               e expression                                      |
|                                                                 |
|   resultat : float correspondant a l'expression evaluee         |
------------------------------------------------------------------*)

let rec evalue_expression env e =
  let evalue = evalue_expression env in
  match e with
    
    |Plus(g,d) -> (evalue g) +. (evalue d)
    
    |Moins(g,d) -> (evalue g) -. (evalue d)
    
    |Div(g,d) -> (evalue g) /. (evalue d)
    
    |Mult(g,d) -> (evalue g) *. (evalue d)
    
    |Var(nom) -> 
       let (ENV(defliste,varliste))=env in
       valeur_de_var varliste nom
    
    |Const(valeur) -> valeur
    
    |Cosinus(degre) -> cos (rad_of_deg (evalue degre))
    
    |Sinus(degre) -> sin (rad_of_deg  (evalue degre))
    
    |Tangente(degre) -> tan (rad_of_deg (evalue degre)) ;;


(*.................................................................
. Test : #evalue_expression (ENV([],[]))                           .
.                          (Moins (Const 0.0, Const 90.0));;      .
.        - : float = -90.0                                        .
..................................................................*)



(*---------------------------------------------------------------- 
|   fonction evalue_condition : environnement -> test -> bool     |
|                                                                 |
|   semantique : donne la valeur booleenne d'un test pour un      |
|                environnement donne.                             |
|                                                                 |
|   arguments : env environnement                                 |
|               t de type test                                    |
|                                                                 |
|   resultat : booleen correspondant a la valeur du test          |
------------------------------------------------------------------*)

let rec evalue_condition env t =
  let evalue = evalue_condition env in
    match t with

      |Equal(g,d) -> (evalue_expression env g) =. (evalue_expression env d)
      
      |InfEq(g,d) -> (evalue_expression env g) <=. (evalue_expression env d)
      
      |And(tg,td) -> (evalue tg)&&(evalue td)
      
      |Or(tg,td) -> (evalue tg)||(evalue td)
      
      |Not(test) -> not (evalue test) ;;

(*.................................................................
. Test : #evalue_condition (ENV([],[("n"),1.]))                   .
.                          (Equal(Var("n"),Const(1.)));;          .
.        - : bool = true                                          .
..................................................................*)



(*---------------------------------------------------------------- 
|  fonction extraire_proc : (definition list -> string )          |
|                           -> (string list * instruction list)   |
|                                                                 |
|  semantique : extraction de la liste d'arguments et du bloc     |
|               d'une procedure a partir de son nom et d'une      |
|               liste de procedures                               |
|                                                                 |
|  arguments : l : liste de defs de procedures                    |
|              nom : nom de la procedure                          |
|                                                                 |
|  resultat : couple(liste d'arguments,bloc d'instruction)        |
------------------------------------------------------------------*)

let rec extraire_proc l nom =
  match l with
      
    |[] -> failwith "CALL sur une procedure indefinie"
    
    |(PROC(name,listearg,bloc))::q when name=nom -> (listearg,bloc)
    
    |t::q -> extraire_proc q nom ;;

(*.................................................................
. Test : #extraire_proc [(PROC("proc1",["n"],[REMPLIR]));         .
.                        (PROC("proc2",["p"],[ARRETREMPLIR]))]    .
.                       "proc1";;                                 .
.        - : string list * instruction list = ["n"], [REMPLIR]    .
..................................................................*)



(*----------------------------------------------------------------        
|  nouvelle_varliste : (environnement -> 'a list -> expr list)    |
|                      -> ('a * float) list                       |
|                                                                 |
|  semantique : liste de couples (variable,valeur) a partir de    |
|               la liste des variables et celle des valeurs       |
|                                                                 |
|  arguments : env : environnement                                |
|              lnom : liste des noms des variables                |
|              lvar : liste des valeurs des variables             |
|                                                                 |
|  resultat : liste de couples (variable,valeur)                  |
------------------------------------------------------------------*)   
       
let rec nouvelle_varliste env lnom lval = 
  match (lnom,lval) with
    
    |([],[]) -> []
    
    |(param::q1,valeur::q2) -> 
       (param,(evalue_expression env valeur))::(nouvelle_varliste env q1 q2)
    
    |_ -> failwith "CALL : erreur dans le nbre de parametres" ;;

(*.................................................................
. Test : #nouvelle_varliste (ENV([],[]))                          .
.                           ["a";"b"]                             .
.                           [Const(5.);Const(7.)];;               .
.         - : (string * float) list = ["a", 5.0; "b", 7.0]        .
..................................................................*)



(*----------------------------------------------------------------
|  execute_instruction : (environnement -> instruction -> etat)   | 
|                        -> etat                                  |
|                                                                 |
|  semantique : execute une instruction et renvoie le nouvel      |
|               etat apres execution.                             | 
|                                                                 |
|  arguments : env : environnement                                |
|              inst : instruction                                 |
|              etat : etat du systeme                             |
|                                                                 |
|  resultat : etat apres execution de l'instruction               |
------------------------------------------------------------------*)   

let rec execute_instruction env inst etat =
  match inst with
    
    |SI(test,alors,sinon) -> 
       if (evalue_condition env test) then
	 (execute_instruction env (LISTE(alors)) etat)
       else
	 (execute_instruction env (LISTE(sinon)) etat);
    
    |REPETE(e,liste) -> 
       let i = (int_of_float (evalue_expression env e))in
	 if (i <= 0) then
	   etat
         else begin
	      let etat2= (execute_instruction env (LISTE(liste)) etat) in
	      execute_instruction env 
                                  (REPETE((Const(float_of_int (i-1)),liste))) 
                                  etat2;
	      end

    |APPEL(nom,liste_val) -> 
       let (ENV(liste_proc,varliste))=env in
       let (liste_arguments,bloc)=(extraire_proc liste_proc nom) in
       let newvarliste = nouvelle_varliste env liste_arguments liste_val in  
       let nouvel_env = (ENV(liste_proc,newvarliste)) in
       (execute_instruction nouvel_env
	                     (LISTE(bloc))
	                      etat)

    |ROTATION(angle)-> 
       let rad=(rad_of_deg (evalue_expression env angle)) in
	 {abs = etat.abs ; 
	  ord = etat.ord ; 
	  ang = (etat.ang +. rad); 
	  remp = etat.remp ; 
	  ox = etat.ox ; 
	  oy = etat.oy }  

    |DEPLACE(e) -> 
       let l=(evalue_expression env e) in
       let x=etat.abs and y=etat.ord and angle=etat.ang in
       let newx=(x+.(l*.cos(angle))) and newy=(y+.(l*.sin(angle))) in
       let nouveletat= {abs=newx ; 
			ord=newy ; 
			ang=angle ; 
			remp=etat.remp ; 
			ox=etat.ox ; 
			oy = etat.oy} 
       in
	 if ((etat.remp) && ((x<>etat.ox) || (y<>etat.oy))) then 
	   begin
	   ftriangle (etat.ox,etat.oy) (x,y) (newx,newy); 
	   nouveletat
	   end
	 else 
	   begin
	   flineto (newx,newy); 
	   nouveletat;
	   end

    |SAUT(e) -> 
       let l=(evalue_expression env e) in
       let x=etat.abs and y=etat.ord and angle=etat.ang in
       let newx=(x+.(l*.cos(angle))) and newy=(y+.(l*.sin(angle))) in
	 fmoveto (newx,newy);
	 {abs=newx ; 
	  ord=newy ; 
	  ang=angle; 
	  remp=etat.remp ; 
	  ox=etat.ox ; 
	  oy = etat.oy }
	   
    |COULEUR(er,eg,eb) -> 
       let r =(evalue_expression env er) and
	   g =(evalue_expression env eg) and
	   b =(evalue_expression env eb) in
	 fcolor r g b;
	 etat

    |REMPLIR -> 
       {abs=etat.abs ; 
	ord=etat.ord ; 
	ang=etat.ang ; 
	remp=true ; 
	ox=etat.abs ; 
	oy=etat.ord }
    
    |ARRETREMPLIR -> 
       {abs=etat.abs ; 
	ord=etat.ord ; 
	ang=etat.ang ; 
	remp=false ; 
	ox=etat.ox ; 
	oy=etat.oy }

    |LISTE(bloc) -> 
       begin
         match bloc with
 	   |[] -> etat
	   |t::q->let new_etat=(execute_instruction env t etat) in
		    execute_instruction env (LISTE(q)) new_etat
       end ;; 

(*.................................................................
. Test : consulter le rapport.                                    .
..................................................................*)			



(*---------------------------------------------------------------- 
|  Procedure execute_programme : programme -> unit                |
|                                                                 |
|  semantique : Produit un dessin a partir d'un programme LOGO    |
|                                                                 |
|  arguments : prog : programme                                   |
------------------------------------------------------------------*)

let execute_programme prog =
  fmoveto (0.,0.);
  fcolor 0. 0. 0.;
  let etat={abs=0.;ord=0.;ang=0.;remp=false;ox=0.;oy=0.} in
  let (PROG(defliste,mainbloc))=prog in
  let env=ENV(defliste,[]) in
  let etat_final=(execute_instruction env (LISTE(mainbloc)) etat) in
  () ;;

(*.................................................................
. Test : consulter le rapport.                                    .
..................................................................*)
